/**********************************************************************************
 *                                                                                *
 * dstrdf_aasen: Factorization routine for Aasen blocked triangular               * 
 *               tridiagonalization                                               *
 *                                                                                *
 * See 'driver.c' for a propoer usage example.                                    *
 *                                                                                *
 *                                                                                *
 *  Gil Shklarski and Sivan Toledo, May 2007                                      *
 *                                                                                *
 **********************************************************************************/


#include <math.h>

#define min(x,y) (((x) < (y))? (x) : (y))


/* BLAS Headers */
void dcopy_(int * N, double * DX, int * INCX, double * DY,int * INCY);
void dscal_(int * N, double * DA, double * DX, int *INCX);
void daxpy_(int * N, double * DA, double * DX, int * INCX, double * DY, int * INCY);
void dgemv_(char * TRANS,int * M, int * N,double * ALPHA, double * A,int * LDA,
            double * X, int * INCX, double * BETA, double * Y, int * INCY);
void dgemm_(char * TRANSA, char * TRANSB, int * M,int * N,int *K,
	    double * ALPHA, double * A, int * LDA,
	    double * B, int * LDB, 
	    double * BETA, double * C, int * LDC);
int  idamax_(int * N,  double * DX, int * INCX);
void dswap_(int * N, double * DX, int * INCX, double * DY,int * INCY);


/* 
   N    = (INPUT) matrix order.
   NB   = (INPUT) number of columns in a partition
   A    = (INPUT/OUTPUT) matrix stored colomnwise
   LDA  = (INPUT) A's leading dimension
   IPIV = (OUTPUT) pivoting index switch array (of size N)
   buf  = (INPUT) temporary buffer (of size NB*N)
   e1,tmp,up_col,up_row = (INPUT) temporary buffers (of size N each).
   UP_BLOCK_FCT = (INPUT)Controls blocking in the update phase, choose '1'
   INFO = (INPUT)For Future Use

*/
void dstrdf_aasen( int * N, double * A, int *LDA, int * IPIV, 
		   double * buf, double * e1, double * tmp, 
		   double * up_col, double * up_row,
		   int * NB, int * UP_BLOCK_FCT, int * INFO ) {

  int J,J2,J3,tB,j,i,k;
  int BL;
  int Jb;
  const int n = *N;
  const int lda = *LDA;
  int nb = *NB;
  int nblocks = (int)ceil((double)n/(double)nb);
 char charN = 'N', charT = 'T';
  int _M,_N,_K;
  double minus_one = -1.0, one = 1.0;
  int one_int = 1;

  double alpha;
  int m1,m2,max_ind;
  double max_tmp;
  double tp;

  int update_block_factor = *UP_BLOCK_FCT;


  if (n <= 2) {
    if ((nblocks) != 1) 
      *INFO = 1;
    else {
      IPIV[1] = 1;
      IPIV[2] = 2;
      *INFO = 0;
    }
    return;
  } 

  J = 0;

  for (i=1;i<=(n-1);i++)
    e1[i] = 0;
  
  for (i=1;i<=n;i++)
    IPIV[i] = i;

  for (BL=1;BL<=(nblocks);BL++) {
    Jb = J+1;
    
   
    for (J=Jb; J<=min(n,Jb+nb-1); J++) {
      j = J-Jb+1;

    
      _M = n-J+1;
      _N = j-1;

      dcopy_(&_M,A+J+(J-1)*lda,&one_int,buf+J+(j-1)*lda,&one_int);


      dgemv_(&charN, &_M, &_N, &minus_one, 
	     buf+J, LDA, A+J+(Jb-1)*lda,LDA,
	     &one, buf+J+(j-1)*lda,&one_int);
      
      dcopy_(&_M,buf+J+(j-1)*lda,&one_int,tmp+1,&one_int);
 

      if (J>1) {
	alpha = -A[J-1+(J-1)*lda]; 
	daxpy_(&_M,&alpha,A+J+(J-2)*lda,&one_int,tmp+1,&one_int);
      }

    
      A[J+(J-1)*lda] = tmp[1];

      if (J<n) {
	alpha = -tmp[1];
	tmp[1] = 0;
	_M = n-J;
	daxpy_(&_M,&alpha,e1+1,&one_int,tmp+2 ,&one_int);	

	
	max_ind = idamax_(&_M,tmp+2,&one_int);
	max_ind = max_ind + 1;
	max_tmp = tmp[max_ind];
	
	if ((max_ind != 1) && (max_tmp != 0)) {
	  m1 = 2;
	  m2 = max_ind;

	  tp = e1[1];
	  e1[1] = e1[m2-1];
	  e1[m2-1] = tp;


	  tmp[m2] = tmp[m1];
	  tmp[m1] = max_tmp;

	  m1 = m1+J-1;
	  m2 = m2+J-1;

	  _M = m2-m1;
	  dcopy_(&_M,
		 A+m1+(m1-1)*lda,&one_int,
		 A+m1+(m2-1)*lda,&one_int);


	  dcopy_(&_M,
		 A+m2+(m1-1)*lda,LDA,
		 A+m1+(m1-1)*lda,&one_int);
    
	  if (BL > 1) {	
	    alpha = up_row[m2-Jb+1];
	    /* Y = alpha*X+Y */
	    daxpy_(&_M,
		   &alpha,
		   up_col+m1-Jb+1, &one_int,
		   A+m1+(m1-1)*lda,&one_int);
	    alpha = -up_col[m2-Jb+1];
	    daxpy_(&_M,
		   &alpha,
		   up_row+m1-Jb+1, &one_int,
		   A+m1+(m1-1)*lda,&one_int);
	  }
	  
	    
	      
	  _N = n-m2+1;
	  dswap_(&_N, A+m2+(m1-1)*lda,&one_int,A+m2+(m2-1)*lda,&one_int);
	  
	  _M = m1-Jb;
	  dswap_(&_M, buf+m1,LDA,buf+m2,LDA);

	  IPIV[m1] = m2;


	  _N = m1-2-Jb+1;
	  if (_N > 0)
	    dswap_(&_N,A+m1+(Jb-1)*lda,LDA,A+m2+(Jb-1)*lda,LDA);


	  tp = A[m2+(m1-1)*lda];
	  A[m2+(m1-1)*lda] = A[m1+(m1-1)*lda];
	  A[m1+(m1-1)*lda] = tp;
	
	  _M = m2-1-m1;
	  dcopy_(&_M,A+(m1+1)+(m2-1)*lda,&one_int,A+m2+m1*lda,LDA);
	  if (BL > 1) {
	    alpha = -up_row[m1-Jb+1];
	    /* Y = alpha*X+Y */
	    daxpy_(&_M,
		   &alpha,
		   up_col+m1+1-Jb+1,&one_int,
		   A+m2+m1*lda,LDA);
	    alpha = up_col[m1-Jb+1];
	    daxpy_(&_M,
		   &alpha,
		   up_row+m1+1-Jb+1,&one_int,
		   A+m2+m1*lda,LDA); 

	  }


	  A[m2+(m2-1)*lda] = A[m1+(m2-1)*lda];

	  if (BL > 1) {
	    tp = up_row[m1-Jb+1];
	    up_row[m1-Jb+1] = up_row[m2-Jb+1];
	    up_row[m2-Jb+1] = tp;

	    tp = up_col[m1-Jb+1];
	    up_col[m1-Jb+1] = up_col[m2-Jb+1];
	    up_col[m2-Jb+1] = tp;


	  }

	}

	for (k=(J+1); k <= n; k++) 
	  A[(J-1)*lda+k] = e1[k-J];



	A[J+J*lda] = tmp[2];

	if (tmp[2] != 0.0) {
	  _M = n-J-1;
	  alpha = 1/tmp[2];

	  dcopy_(&_M,tmp+3,&one_int,e1+1,&one_int);
	  dscal_(&_M,&alpha,e1+1,&one_int);

	} else {
	  for (i=1; i<= (n-J-1); i++)
	    e1[i] = 0;
	}
	/* for (i=1;i<=_M;i++)
	  mexPrintf("XXX E1 %e\n",e1[i]);
	*/
      }

    }
    J=J-1; /* MATLAB - C difference */
    if (J < n) {
      

      for (J2 = J; J2<= n; J2+=(update_block_factor*nb)) {
	tB = min(update_block_factor*nb,n-J2);

	/* We compute an extra NB-by-NB triangle to use only BLAS 3 */
	_M = n-J2;
	_N = tB;
	_K = min(nb,n-Jb+1);

	dgemm_(&charN, &charT, &_M, &_N, &_K,
	       &minus_one, buf+J2+1, LDA,
	       A+J2+1+(Jb-1)*lda, LDA,
	       &one,  A+J2+1+J2*lda, LDA);
	
	/* 
	   //The other option for computing is BLAS 3 for rectangular blocks and
	   // repeated BLAS 2 for triangular blocks
	_N = min(nb,n-Jb+1);
	for (J3 = (J2+1); J3 <= (J2+tB); J3++) {
	  _M = J2+tB-J3+1;
	  dgemv_(&charN,&_M,&_N,
		 &minus_one,buf+J3, LDA,
		 A+J3+(Jb-1)*lda, LDA,
		 &one, A+J3+(J3-1)*lda,&one_int);
	}
	_M = n-J2-tB;
	_N = tB;
	_K = min(nb,n-Jb+1);

	dgemm_(&charN, &charT, &_M, &_N, &_K,
	       &minus_one, buf+J2+tB+1, LDA,
	       A+J2+1+(Jb-1)*lda, LDA,
	       &one,  A+J2+tB+1+J2*lda, LDA);


	*/

      }

      _M = n-J;
      dcopy_(&_M,A+J+1+(J-1)*lda,&one_int,up_col+1,&one_int);

      _M = n-J-1; 
      up_row[1] = 1;
      if (_M > 0) 
	dcopy_(&_M,e1+1,&one_int,up_row+2,&one_int);
      _M++;
      dscal_(&_M,A+J+J*lda,up_row+1,&one_int);

    }


  }
  *INFO = 0;

}
